#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#if STDC_HEADERS
#  include <stdlib.h>
#else
#  if HAVE_MALLOC_H
#    include <malloc.h>
#  endif
#endif

#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "proto_decl.h"
#include "browser_init.h"
#include "global_parse.h"
#include "result_types.h"
#include "file_dict.h"
#include "proc_dict.h"
#include "class_dict.h"
#include "global_dict.h"
#include "browser_pars.h"
#include "parse.h"
#include "browser_eve.h"
#include "page.h"

#include "BR_Global.h"
#include "BR_Interface.h"
 
extern Browser br;

// LF Avril 2020
extern int chdir (char *);
extern void class_eraze_file (char * filename); 
extern BrowserError add_class_decl(char *, int, unsigned, unsigned);
extern BrowserError add_proc_impl(char *, ProcType, int, unsigned, unsigned);
extern void class_eraze_marked_files();
extern void browser_show_all();
extern void proc_eraze_file(char *);
extern void global_eraze_file(char *);

#if !HAVE_STRDUP
extern FCT (char*, strdup, (char*) );
#endif

/*-------------------------------------------------------------------------
*/
static ScopeType compute_scope(infos)
    BR_ULong infos;
{
  return (infos & BR_INFO_PUBLIC)
    ? PUBLIC_SCOPE
    : ((infos & BR_INFO_PROTECTED)
       ? PROTECTED_SCOPE
       : ((infos & BR_INFO_PRIVATE) ? PRIVATE_SCOPE : UNKNOWN_SCOPE));
}

static int yaerror;
static char * filename;

void
browser_handler(str_prefix, str_syntax_unit, infos, line_num)
    BR_Str*  str_prefix;
    BR_Str*  str_syntax_unit;
    BR_ULong infos;
    BR_ULong line_num;
{
  static BR_Str* erroned_prefix = 0;
  
  if (infos & BR_INFO_ERROR) {
    if (OpVerbose()) {
      if (! yaerror)
	fprintf(stderr, "Browser : parse error in %s line(s) %lu",
		filename, line_num);
      else
	fprintf(stderr, " %lu", line_num);
    }
    yaerror = 1;
    return;
  }
  if (str_prefix) {
    if (*str_prefix == '?')
      return;
    if (erroned_prefix) {
      BR_Str * s1, *s2;
      
      for (s1 = str_prefix, s2 = erroned_prefix; (*s1)&&(*s2); s1 += 1, s2 += 1)
	if (*s1 != *s2)
	  break;
      
      if (! *s1) {
	if (! *s2)
	  /* member of redefined class */
	  return;
	/* out of redefined class */
	free(erroned_prefix);
	erroned_prefix = 0;
      }
      else if (! *s2) {
	if (*s1 == ':')
	  /* sub class of redefined class */
	  return;
	/* out of redefined class */
	free(erroned_prefix);
	erroned_prefix = 0;
      }
    }
    if (((infos & BR_INFO_CLASS) && (infos & BR_INFO_DEF)) ||
	(infos & (BR_INFO_STRUCT | BR_INFO_UNION | BR_INFO_TYPEDEF))) {
      char * n = (char *)
	malloc(strlen(str_prefix) + strlen(str_syntax_unit) + 3);
      
      sprintf(n, "%s::%s", str_prefix, str_syntax_unit);
      if ((*str_syntax_unit == '?') ||
	  (add_class_decl(n, line_num, (infos & BR_INFO_TYPEDEF) ? 1 : 0, 0) ==
	   BERROR))
	erroned_prefix = n;
      else
	free(n);
    }
    else if (infos & BR_INFO_PARENT)
      add_parent(str_prefix, str_syntax_unit, line_num, compute_scope(infos));
    else if (infos & BR_INFO_PROC) {
      if (infos & BR_INFO_NAMESPACE_ELEMENT) {
	/* a function in a namespace */
	if (infos & BR_INFO_DEF) {
	  char * n = (char *)
	    malloc(strlen(str_prefix) + strlen(str_syntax_unit) + 3);
	  
	  sprintf(n, "%s::%s", str_prefix, str_syntax_unit);
	  add_proc_impl(n, C_PROC, line_num,
			((infos & BR_INFO_STATIC) ? STATIC_PROC : 0) +
			((infos & BR_INFO_TEMPLATE) ? TEMPLATE_PROC : 0),
			0);
	  free(n);
	}
      }
      else {
	if ((!(infos & BR_INFO_DEF)) ||
	    (infos & (BR_INFO_PUBLIC | BR_INFO_PROTECTED | BR_INFO_PRIVATE))) {
	  int decl = ((infos >> 9) & 7)	/* private protected public */
	    | ((infos >> 12) & 0x3f8)	/* inline constructor .. pure */
	      | ((infos >> 14) & 0x400);	/* friend */
	  
	  add_method_decl(str_prefix, str_syntax_unit, compute_scope(infos),
			  decl, line_num, 0);
	}
	if (infos & BR_INFO_DEF) {
	  add_method_impl(str_prefix, str_syntax_unit, line_num, 0);
	}
      }
    }
    else if (infos & BR_INFO_DATA) {
      if (infos & BR_INFO_NAMESPACE_ELEMENT) {
	if (infos & BR_INFO_DEF) {
	  /*int decl = (infos >> 10) & 8;	 static */
	  char * n = (char *)
	    malloc(strlen(str_prefix) + strlen(str_syntax_unit) + 3);
	  
	  sprintf(n, "%s::%s", str_prefix, str_syntax_unit);
	  add_global_impl(n,
			  (infos & BR_INFO_STATIC) ? 1 : 0,
			  line_num, 0);
	  free(n);
	}
      }
      else {
	int decl;
	
	decl = ((infos >> 9) & 7)		/* private protected public */
	  | ((infos >> 10) & 8);	/* static */
	add_attribute_decl(str_prefix, str_syntax_unit, compute_scope(infos),
			   decl, line_num, 0);
      }
    }
  }
  else if (*str_syntax_unit == '?')
    ;
  else if (infos & BR_INFO_PROC) {
    if (infos & BR_INFO_DEF)
      add_proc_impl(str_syntax_unit, C_PROC, line_num,
		    ((infos & BR_INFO_STATIC) ? STATIC_PROC : 0) +
		    ((infos & BR_INFO_INLINE) ? INLINE_PROC : 0) +
		    ((infos & BR_INFO_TEMPLATE) ? TEMPLATE_PROC : 0),
		    0);
  }
  else if (infos & BR_INFO_DATA) {
    if (infos & BR_INFO_DEF)
      add_global_impl(str_syntax_unit, (infos & BR_INFO_STATIC) ? 1 : 0,
		      line_num, 0);
  }
  else if (((infos & BR_INFO_CLASS) && (infos & BR_INFO_DEF)) ||
	   (infos & (BR_INFO_STRUCT | BR_INFO_UNION | BR_INFO_TYPEDEF))) {
    /* new toplevel class */
    if (erroned_prefix)
      free(erroned_prefix);
    erroned_prefix =
      (add_class_decl(str_syntax_unit, line_num,
		      ((infos & BR_INFO_TYPEDEF) ? TYPEDEF_TYPE : 0) +
		      ((infos & BR_INFO_TEMPLATE) ? TEMPLATE_TYPE : 0),
		      0)
       == BERROR)
	? (BR_Str*) strdup(str_syntax_unit) : (BR_Str*) 0;
  }
}


/*-------------------------------------------------------------------------
*/

int browser_yyparse(fn)
    char * fn;
{
  yaerror = 0;
  filename = fn;

  if ((! br.mode->pp_flag) || (! br.mode->pp_path)) {
    BR_parse_file(fn, browser_handler, 0);
  }
  else if (! br.mode->pp_options) {
    static char* filter_cmd[] = { BR_Null, BR_Null, BR_Null };

    filter_cmd[0] = br.mode->pp_path;
    filter_cmd[1] = fn;
    BR_parse_file(fn, browser_handler, filter_cmd);
  }
  else {
    char ** filter_cmd;
    char * options;
    char * p;
    unsigned nentry = 3;
    
    options = strdup(br.mode->pp_options);
    p = strtok(options, " \t");
    while (p) {
      nentry += 1;
      p = strtok(0, " \t");
    }
    
    filter_cmd = (char **) malloc(nentry * sizeof(char *));
    filter_cmd[0] = br.mode->pp_path;
    
    strcpy(options, br.mode->pp_options);
    p = strtok(options, " \t");
    nentry = 1;
    
    while (p) {
      filter_cmd[nentry] = strdup(p);
      p = strtok(0, " \t");
      nentry += 1;
    }
    filter_cmd[nentry] = fn;
    filter_cmd[nentry + 1] = BR_Null;
  
    BR_parse_file(fn, browser_handler, filter_cmd);
    
    while (--nentry) {
      free(filter_cmd[nentry]);
    }
    free(filter_cmd);
    free(options);
  }
  
  if (yaerror && OpVerbose())
    fputc('\n', stderr);

  return 0;		/* yaerror */
}

int yywrap() { return 1; }

/*-------------------------------------------------------------------------
   fonctions Smac
*/

void browser_set_pp(modename, path)
    char * modename;
    char * path;
{
  Mode * mode = GetMode(modename);
  
  if (! mode)
    return;
  
  if (mode->pp_path)
    free(mode->pp_path);
  
  mode->pp_path = strdup(path);
}

void browser_set_pp_options(modename, options)
    char * modename;
    char * options;
{
  Mode * mode = GetMode(modename);
  
  if (! mode)
    return;
  
  if (mode->pp_options)
    free(mode->pp_options);
  if (options) {
    mode->pp_options = strdup(options);
    if (! mode->pp_flag) {
      mode->pp_flag = 1;
      RefreshPpButton();
    }
  }
  else {
    mode->pp_options = 0;
    if (mode->pp_flag) {
      mode->pp_flag = 0;
      RefreshPpButton();
    }
  }
}

void browser_add(s, rec)
    char * s;
    int rec;
{
  if ((s != 0) && (strlen (s) != 0 )) {
    struct stat st;
    int old_rec = br.parse_flag;
    char pathname [MAXPATHLEN+2];
    char * here;
    extern char  *getcwd();
    
    br.parse_flag = (rec) ? RECURSIVE : NO_RECURSIVE;
#if HAVE_GETCWD
    (void) getcwd ( pathname, MAXPATHLEN );
#else
    if ( getwd ( pathname ) == 0 ) {
      (void) fprintf ( stderr, "Can't getwd\n" );
      return;
    }
#endif
    here = (char *) malloc(strlen(pathname) + 1);
    if (! here) return;
    (void) strcpy (here, pathname);
    
    (void) stat ( s, &st );
    if ( S_ISDIR(st.st_mode) != 0 ) {
      (void) chdir(s);
#if HAVE_GETCWD
      (void) getcwd ( pathname, MAXPATHLEN );
#else
      if ( getwd ( pathname ) == 0 ) {
	(void) fprintf ( stderr, "Can't getwd\n" );
	(void) chdir(here);
	free(here);
	return;
      }
#endif
      LoadDir ( pathname );
    }
    else if (S_ISREG(st.st_mode) && GoodSuffix (s)) {
      char * p = strrchr(s, '/');
      
      if (! p)
	sprintf(pathname + strlen(pathname), "/%s", s);
      else if (p == pathname)
	sprintf(pathname, "/%s", s);
      else {
	char c = *p;
	
	*p = 0;
	(void) chdir(s);
	*p = c;
	
	if (*s == '/')
	  strcpy(pathname, s);
	else
	  sprintf(pathname + strlen(pathname), "/%s", p+1);
      }
      parse_file (pathname); 
    }
    br.parse_flag = old_rec;
    (void) RefreshBrowserInfos ();
    
    (void) chdir(here);
    free(here);
  }  
}

char * browser_class_file(class_name, pline)
    char * class_name;
    int * pline;
{
  ClassRec* class = find_class(class_name);
  
  if (class) {
    *pline = class->_decl_line;
    return class->_decl_file->_name;
  }
  else
    return 0;
}


char * browser_class_flags(class_name)
    char * class_name;
{
  ClassRec* class = find_class(class_name);
  
  if (class) {
    return class->_name - CLASS_PLENGTH;
  }
  else
    return 0;
}

static char * browser_select_x(prefix, t, w_x, x_plength)
    char * prefix;
    char **t;
    unsigned w_x;
    unsigned x_plength;
{
  if (t) {
    unsigned plen = strlen(prefix);
    int i = 1;
    
    DisplayBrowserIfNeverMap();
    
    while (*t != 0) {
      if (! HIDED(*t)) {
	int d = strncmp(prefix, *t, plen);
	
	if (d == 0) {
	  BWin *b_x = &br.tbw[w_x];
	  
	  UpdateTitle (*t - x_plength);
	  /* i = no ligne dans la liste des classes */
	  UpdateTextItem ( b_x -> text, b_x -> select );
	  GotoLineNumber ( b_x -> text, i );
	  CurrentLineToTop ( b_x -> text );
	  UpdateTextItem ( b_x -> text, b_x -> select );
	  SelectItem ( b_x, 5, 5, w_x );
	  
	  return *t;
	}
	if (d < 0)
	  return 0;
	i += 1;
      }
      t += 1;
    }
  }
  
  return 0;
}

int browser_select_class(prefix)
    char * prefix;
{
  char * s = 
    browser_select_x(prefix, br.class_save, W_CLASS, CLASS_PLENGTH);
  
  return (s == 0) ? 0 : strlen(s);
}

int browser_select_method(prefix)
    char * prefix;
{
  char * s = 
    browser_select_x(prefix, br.methods_save, W_METHOD, METHOD_PLENGTH);
  
  return (s == 0) ? 0 : strchr(s, '(') - s;
}

int browser_select_attribute(prefix)
    char * prefix;
{
  char * s = 
    browser_select_x(prefix, br.attributes_save,
		     W_ATTRIBUTES, ATTRIBUTE_PLENGTH);
  
  return (s == 0) ? 0 : strlen(s);
}

int browser_select_proc(prefix)
    char * prefix;
{
  char * s = 
    browser_select_x(prefix, br.procs_save, W_PROC, PROC_PLENGTH);
  
  return (s == 0) ? 0 : strchr(s, '(') - s;
}

int browser_select_global(prefix)
    char * prefix;
{
  char * s = 
    browser_select_x(prefix, br.globals_save, W_GLOBALS, GLOBAL_PLENGTH);
  
  return (s == 0) ? 0 : strlen(s);
}

static int browser_x_entry(prefix, save)
    char * prefix;
    char ** save;
{
  if (save) {
    unsigned plen = strlen(prefix);
    char ** p;
 
    for (p = save; *p; p += 1) {
      int d = strncmp(*p, prefix, plen);
      
      if (d == 0)
	return p - save;
      if (d > 0)
	return -1;
    }
  }
  
  return -1;
}

int browser_class_entry(prefix)
    char * prefix;
{
  return browser_x_entry(prefix, br.class_save);
}

int browser_function_entry(prefix)
    char * prefix;
{
  return browser_x_entry(prefix, br.procs_save);
}

int browser_global_entry(prefix)
    char * prefix;
{
  return browser_x_entry(prefix, br.globals_save);
}

int browser_file_entry(prefix)
    char * prefix;
{
  return browser_x_entry(prefix, br.files_save);
}

static char * browser_x(rang, save)
    int rang;
    char ** save;
{
  if (save)
    for (; *save; save += 1, rang -= 1)
      if (! rang)
	return *save;
  
  return 0;
}

char * browser_class(n)
    int n;
{
  return browser_x(n, br.class_save);
}

char * browser_function(n)
    int n;
{
  return browser_x(n, br.procs_save);
}

char * browser_global(n)
    int n;
{
  return browser_x(n, br.globals_save);
}

char * browser_file(n)
    int n;
{
  return browser_x(n, br.files_save);
}

/* ****
   DUMP
   **** */
   
#define DUMP_VERSION1 1
#define DUMP_VERSION2 2

#define TYPE_FIN		0
#define TYPE_CLASS_DECL		1
#define TYPE_PARENT_DECL	2
#define TYPE_METHOD_DECL	3
#define TYPE_METHOD_IMPL	4
#define TYPE_ATTRIBUT		5
#define TYPE_INFO_PROC		6
#define TYPE_INFO_DATA		7

extern FCT(void, DisplayWMessage ,(char *, char *, int));

/* to memorize data defined in a file */

typedef struct ClassRecList {
  ClassRec * elt;
  struct ClassRecList * next;
} ClassRecList;
  
typedef struct ProcRecList {
  ProcRec * elt;
  struct ProcRecList * next;
} ProcRecList;
  
typedef struct MethodRecList {
  MethodRec * elt;
  struct MethodRecList * next;
} MethodRecList;
  
typedef struct GlobalRecList {
  GlobalRec * elt;
  struct GlobalRecList * next;
} GlobalRecList;

typedef struct FileData {
  ClassRecList * lclass;
  ProcRecList * lproc;
  MethodRecList * lmethod;	/* implementation part for not inline methods */
  GlobalRecList * lglobal;
} FileData;

/* link each file and its functions */
void proc_memo_file()
{
  ProcRec ** pdict;
  
  for (pdict = proc_dict;
       pdict != &proc_dict[PROC_DICT_SIZE];
       pdict += 1) {
    ProcRec * current_proc;
    
    for (current_proc = *pdict;
	 current_proc != Null;
	 current_proc = current_proc->_next) {
      if (current_proc->_impl_file != 0) {
	FileData * fd = (FileData *) current_proc->_impl_file->_name;
	ProcRecList * cell = (ProcRecList *) malloc(sizeof(ProcRecList));
	
	cell->next = fd->lproc;
	cell->elt = current_proc;
	fd->lproc = cell;
      }
    }
  }
}

/* link each file and its globals */
void global_memo_file()
{
  GlobalRec ** pdict;
  
  for (pdict = global_dict;
       pdict != &global_dict[GLOBAL_DICT_SIZE];
       pdict += 1) {
    GlobalRec * current_global;
    
    for (current_global = *pdict;
	 current_global != Null;
	 current_global = current_global->_next) {
      if (current_global->_impl_file != 0) {
	FileData * fd = (FileData *) current_global->_impl_file->_name;
	GlobalRecList * cell = (GlobalRecList *) malloc(sizeof(GlobalRecList));
	
	cell->next = fd->lglobal;
	cell->elt = current_global;
	fd->lglobal = cell;
      }
    }
  }
}

/* link each file and its classes and methods */
void class_memo_file()
{
  ClassRec ** pdict;
  
  for (pdict = class_dict;
       pdict != &class_dict[CLASS_DICT_SIZE];
       pdict += 1) {
    ClassRec * current_class;
    
    for (current_class = *pdict;
	 current_class != Null;
	 current_class = current_class->_next) {
      FileRec * current_file = current_class->_decl_file;
      
      if (current_file != Null) {
	FileData * fd = (FileData *) current_file->_name;
	ClassRecList * cell = (ClassRecList *) malloc(sizeof(ClassRecList));
	
	cell->next = fd->lclass;
	cell->elt = current_class;
	fd->lclass = cell;
      }
      
      /* method impl not inline */
      {
	MethodRec * current_method = current_class->_methods_list;
	
	while (current_method) {
	  if ((current_method->_impl_file != Null) &&
	      (current_method->_impl_file != current_file)) {
	    FileData * fd = (FileData *) current_method->_impl_file->_name;
	    MethodRecList * cell = (MethodRecList *) malloc(sizeof(MethodRecList));
	    
	    cell->next = fd->lmethod;
	    cell->elt = current_method;
	    fd->lmethod = cell;
	  }
	  current_method = current_method->_next;
	}
      }
    }
  }
}

/* saves a number */
static void dump_nbr(nbr, fp)
    unsigned nbr;
    FILE * fp;
{
  fputc(nbr >> 24, fp);
  fputc(nbr >> 16, fp);
  fputc(nbr >> 8, fp);
  fputc(nbr, fp);
}

/* saves a string */
static void dump_str(str, fp)
    char * str;
    FILE * fp;
{
  int len = strlen(str) + 1;
  
  dump_nbr(len, fp);
  fwrite(str, len, 1, fp);
}

/* saves functions */
static void save_procs(lproc, fp)
    ProcRecList * lproc;
    FILE * fp;
{
  while (lproc != Null) {
    ProcRec * current_proc = lproc->elt;
    ProcRecList * next = lproc->next;
    
    fputc(TYPE_INFO_PROC, fp);
    dump_str(current_proc->_name, fp);
    dump_nbr(current_proc->_impl_line, fp);
    fputc(current_proc->_decl, fp);
    dump_nbr(current_proc->_hide, fp);
    
    free(lproc);
    lproc = next;
  }
}

/* saves classes */
static void save_classes(lclass, fp, current_file)
    ClassRecList * lclass;
    FILE * fp;
    FileRec * current_file;
{
  while (lclass != Null) {
    ClassRec * current_class = lclass->elt;
    ClassRecList * next = lclass->next;
    
    fputc(TYPE_CLASS_DECL, fp);
    dump_str(current_class->_name, fp);
    dump_nbr(current_class->_decl_line, fp);
    fputc(current_class->_decl, fp);
    dump_nbr(current_class->_hide, fp);
    {
      ParentRec * current_parent = current_class->_parents_list;
      
      while (current_parent) {
	fputc(TYPE_PARENT_DECL, fp);
	dump_str(current_parent->_name, fp);
	dump_nbr(current_parent->_decl_line, fp);
	fputc(current_parent->_scope, fp);
	current_parent = current_parent->_next;
      }
    }
    {
      MethodRec * current_method = current_class->_methods_list;
      
      while (current_method) {
	if (current_method->_decl_file == current_file) {
	  /* si ce n'est pas le cas il y a une double definition
	     de la classe */
	  fputc(TYPE_METHOD_DECL, fp);
	  dump_str(current_method->_name, fp);
	  fputc(current_method->_scope, fp);
	  fputc(current_method->_decl, fp);
	  dump_nbr(current_method->_decl_line, fp);
	  dump_nbr(current_method->_hide, fp);
	  if (current_method->_impl_file == current_file) {
	    fputc(TYPE_METHOD_IMPL, fp);
	    dump_nbr(current_method->_impl_line, fp);
	  }
	  else
	    fputc(TYPE_FIN, fp);
	}
	current_method = current_method->_next;
      }
    }
    {
      AttributeRec * current_attr = current_class->_attributes_list;
      
      while (current_attr) {
	if (current_attr->_decl_file == current_file) {
	  /* si ce n'est pas le cas il y a une double definition
	     de la classe */
	  fputc(TYPE_ATTRIBUT, fp);
	  dump_str(current_attr->_name, fp);
	  fputc(current_attr->_scope, fp);
	  fputc(current_attr->_decl, fp);
	  dump_nbr(current_attr->_decl_line, fp);
	  dump_nbr(current_attr->_hide, fp);
	}
	current_attr = current_attr->_next;
      }
    }
    
    fputc(TYPE_FIN, fp);
    
    free(lclass);
    lclass = next;
  }
}

/* saves methods impl not inline */
static void save_methods(lmethod, fp)
    MethodRecList * lmethod;
    FILE * fp;
{
  while (lmethod != Null) {
    MethodRec * current_method = lmethod->elt;
    MethodRecList * next = lmethod->next;
    
    fputc(TYPE_METHOD_IMPL, fp);
    dump_str(current_method->_class_name, fp);
    dump_str(current_method->_name, fp);
    dump_nbr(current_method->_impl_line, fp);
    dump_nbr(current_method->_hide, fp);
    
    free(lmethod);
    lmethod = next;
  }
}

/* saves globals */
static void save_globals(lglobal, fp)
    GlobalRecList * lglobal;
    FILE * fp;
{
  while (lglobal != Null) {
    GlobalRec * current_global = lglobal->elt;
    GlobalRecList * next = lglobal->next;
    
    fputc(TYPE_INFO_DATA, fp);
    dump_str(current_global->_name, fp);
    fputc(current_global->_static, fp);
    dump_nbr(current_global->_impl_line, fp);
    dump_nbr(current_global->_hide, fp);
    
    free(lglobal);
    lglobal = next;
  }
}

int browser_dump(fn)
    char * fn;
{
  FILE * fp = fopen(fn, "w");
  unsigned nfile = file_count;
  unsigned ifile;
  
  if (! fp) {
    DisplayWMessage("cannot open ", "browser dump", 1);
    DisplayWMessage(fn, 0, 0);
    DisplayWMessage("\n", "browser dump", 0);
    return 0;
  }
  
  fputc(DUMP_VERSION2, fp);
  
  /* save file list and change each _name to reference data defined in each file */
  
  dump_nbr(nfile, fp);
  
  for (ifile = 0; nfile; ifile += 1) {
    FileRec * current_file;
    
    for (current_file = file_dict[ifile];
	 current_file;
	 current_file = current_file->_next) {
      dump_str(current_file->_name, fp);
      current_file->_name = (char *) malloc(sizeof(FileData));
      memset(current_file->_name, 0, sizeof(FileData));
      nfile -= 1;
    }
  }
  
  /* set FileData */
  
  proc_memo_file();
  global_memo_file();
  class_memo_file();
  
  /* dump */
  
  dump_nbr(Hide_Bits, fp);
  nfile = file_count;
  
  for (ifile = 0; nfile; ifile += 1) {
    FileRec * current_file;
    
    for (current_file = file_dict[ifile];
	 current_file;
	 current_file = current_file->_next) {
      FileData * fd = (FileData *) current_file->_name;
      
      nfile -= 1;
      dump_str(((char *) current_file) + FILE_NAME_OFFSET, fp);
      
      /* saves datas */
      save_procs(fd->lproc, fp);
      save_classes(fd->lclass, fp, current_file);
      save_methods(fd->lmethod, fp);
      save_globals(fd->lglobal, fp);
      
      fputc(TYPE_FIN, fp);
      
      free(fd);
      
      /* restore file's name */
      current_file->_name = ((char *) current_file) + FILE_NAME_OFFSET;
    }
  }
  
  fclose(fp);
  return 1;
}


static int restore_nbr(fp)
    FILE * fp;
{
  unsigned char s[4];
  
  if (fread(s, 1, 4, fp) == 0)
    fprintf (stderr, "fread failed\n");
  
  return (((((s[0] << 8) + s[1]) << 8) + s[2]) << 8) + s[3];
}

static char * restore_str(fp)
    FILE * fp;
{
  int len = restore_nbr(fp);
  char * str = (char *) malloc(len);
  
  if (str){
    if (fread(str, len, 1, fp) == 0)
      fprintf (stderr, "fread failed\n");
  }
  
  return str;
}

static int browser_restore_file(fp, version)
    FILE * fp;
    int version;
{
  if (version == DUMP_VERSION2) {
    /* reads files list to mark them to be removed */
    {
      int nfile = restore_nbr(fp);
      
      while (nfile--) {
	char * file_name = restore_str(fp);
	
	mark_remove_file(file_name);
	free(file_name);
      }
    }
    
    /* remove data related to marked files, then remove marked files */
    proc_eraze_marked_files();
    global_eraze_marked_files();
    class_eraze_marked_files();
    remove_marked_files();
  }
  
  /* restore */
  {
    int c;
    
    browser_show_all();
    Hide_Bits = restore_nbr(fp);
    
    while ((c = fgetc(fp)) != EOF) {
      char * file_name;
      int infotype;
      
      ungetc(c, fp);
      
      if ((file_name = restore_str(fp)) == 0) {
	return 0;
      }
      
      if (version == DUMP_VERSION2) {
	if ((parsed_file = create_file(file_name)) == 0) {
	  free(file_name);
	  return 0;
	}
      }
      else {
	/* version 1 */
	parsed_file = find_file(file_name);
	if (parsed_file != Null) {
	  proc_eraze_file(file_name);
	  global_eraze_file(file_name);
	  class_eraze_file(file_name);
	}
	else if ((parsed_file = create_file(file_name)) == 0) {
	  free(file_name);
	  return 0;
	}
      }
      
      free(file_name);
      
      while ((infotype = fgetc(fp)) != TYPE_FIN) {
	switch (infotype) {
	case TYPE_INFO_PROC :
	  {
	    char * name = restore_str(fp);
	    int impl_line = restore_nbr(fp);
	    int decl = fgetc(fp);
	    
	    if (! name)
	      return 0;
	    add_proc_impl(name, C_PROC, impl_line, decl, restore_nbr(fp));
	    free(name);
	  }
	  break;
	case TYPE_CLASS_DECL :
	  {
	    char * class_name = restore_str(fp);
	    
	    if (! class_name)
	      return 0;
	    {
	      int decl_line = restore_nbr(fp);
	      int decl = fgetc(fp);
	      
	      add_class_decl(class_name, decl_line, decl, restore_nbr(fp));
	    }
	    
	    while ((infotype = fgetc(fp)) != TYPE_FIN) {
	      switch (infotype) {
	      case TYPE_PARENT_DECL :
		{
		  char * parent = restore_str(fp);
		  int decl_line = restore_nbr(fp);
		  
		  if (! parent) {
		    free(class_name);
		    return 0;
		  }
		  add_parent(class_name, parent, decl_line, fgetc(fp));
		  free(parent);
		}
		break;
	      case TYPE_METHOD_DECL :
		{
		  char * meth = restore_str(fp);
		  int scope = fgetc(fp);
		  int decl = fgetc(fp);
		  int decl_line = restore_nbr(fp);
		  int hide = restore_nbr(fp);
		  
		  if (! meth) {
		    free(class_name);
		    return 0;
		  }
		  add_method_decl(class_name, meth, scope, decl, decl_line, hide);
		  if (fgetc(fp) == TYPE_METHOD_IMPL)
		    add_method_impl(class_name, meth, restore_nbr(fp), hide);
		  free(meth);
		}
		break;
	      case TYPE_ATTRIBUT :
		{
		  char * attr = restore_str(fp);
		  int scope = fgetc(fp);
		  int decl = fgetc(fp);
		  int decl_line = restore_nbr(fp);
		  
		  if (! attr) {
		    free(class_name);
		    return 0;
		  }
		  add_attribute_decl(class_name, attr, scope, decl,
				     decl_line, restore_nbr(fp));
		  free(attr);
		}
		break;
	      default:
		free(class_name);
		return 0;
	      }
	    }
	    free(class_name);
	  }
	  break;
	case TYPE_METHOD_IMPL :
	  {
	    char * class_name = restore_str(fp);
	    char * meth = restore_str(fp);
	    int impl_line = restore_nbr(fp);
	    
	    if (! class_name)
	      return 0;
	    if (! meth) {
	      free(class_name);
	      return 0;
	    }
	    add_method_impl(class_name, meth, impl_line, restore_nbr(fp));
	    free(class_name);
	    free(meth);
	  }
	  break;
	case TYPE_INFO_DATA :
	  {
	    char * name = restore_str(fp);
	    int staticp = fgetc(fp);
	    int impl_line = restore_nbr(fp);
	    
	    if (! name)
	      return 0;
	    add_global_impl(name, staticp, impl_line, restore_nbr(fp));
	    free(name);
	  }
	  break;
	default:
	  return 0;
	}
      }
    }
  }
  
  return 1;
}

int browser_restore(fn)
    char * fn;
{
  FILE * fp = fopen(fn, "r");
  int result;
  
  if (fp == 0) {
    DisplayWMessage("cannot open file\n", "browser restore", 1);    
    return 0;
  }
  
  switch (fgetc(fp)) {
  case DUMP_VERSION1:
    result = browser_restore_file(fp, DUMP_VERSION1);
    break;
  case DUMP_VERSION2:
    result = browser_restore_file(fp, DUMP_VERSION2);
    break;
  default:
    DisplayWMessage("wrong dump file format\n", "browser restore", 1);    
    fclose(fp);
    return 0;
  }
  
  if (result == 0)
    DisplayWMessage("wrong dump file or not enough place\n", "browser restore", 1);

  parsed_file = Null;
  class_cache = Null;
  fclose(fp);
  
  RefreshBrowserInfos ();
  
  return result;
}

