/* ########################################################################

			       operator.c

   File: operator.c
   Path: /home/fournigault/c/X11/xcoral-2.33/Smac/operator.c
   Description: 
   Created: Tue Feb 21 12:58:52 MET 1995
   Author: Bruno Pages
   Modified: Tue Feb 21 12:58:53 MET 1995
   Last maintained by: Bruno Pages

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Bruno Pages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */



/* Les operateurs */

#include <stdio.h>

#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#include "operator.h"
#include "list.h"
#include "FunctionCall_.h"
#include "Builtin_.h"
#include "Const_.h"
#include "Type_.h"
#include "RefDeref_.h"
#include "type.h"
#include "error.h"
#include "Cast_.h"
#include "cast.h"
#include "const.h"
#include "identifier.h"
#include "array.h"
#include "Var_.h"
#include "operator.h"

#include "decl.h" /**/

/*******************
  
  Operateurs unaires

  *****************/

Instruction * var_reference(mkident)
     List * mkident;
{
  Instruction * ident = (Instruction *) MAKE(mkident);

  if (ident == ERROR)
    return ERROR;
  
  return
    (IsaConst(ident))
      ? ident
      : (IsaLocalVar(ident))
        ? (Instruction *) LocalVarRef__LocalVarRef((LocalVar *)ident)
        : (Instruction *) GlobalVarRef__GlobalVarRef((GlobalVar *)ident);
}

FCT ( Instruction *, make_oper_asterisk,(List *, int)	);

Instruction * make_oper_ampersand(l, del)
    List * l;
    int del;
{
  List * place = (List *) l->info;

  free(l);

  if (del) {
    if (place)
      DELETE(place);
    return 0;
  }
  
  if (place->fct == make_cast) {
    DELETE(place);
    Error("cast unacceptable operand for &", 1);
    return ERROR;
  }

  if (place->fct == make_string)
    return MAKE(place);
  
  if (place->fct == make_identifier)
    return var_reference(place);

  if (place->fct == make_oper_asterisk) {
    /* &*X = X */
    Instruction * result = MAKE((List *) place->info);
    
    free(place);
    
    if (result == ERROR)
      return ERROR;
    else if (! Type__IsaPointer(GetExprType(result))) {
      Delete(result);
      Error("illegal indirection (not a pointer)", 1);
      return ERROR;
    }
    else
      return result;
  }

  if (place->fct == make_array_access)
    return array_reference(place);

  free(place);
  
  Error("illegal reference", 1);
  return ERROR;
}


Instruction * make_oper_asterisk(l, del)
    List * l;
    int del;
{
  if (del) {
    if (l->info)
      DELETE((List *) l->info);
    free(l);
    return 0;
  }
  else {
    Instruction * p = MAKE((List *) l->info);
    Type * type;
    
    free(l);
    
    if (p == ERROR)
      return ERROR;
    
    type = GetExprType(p);
    
    if (! Type__IsaPointer(type)) {
      Delete(p);
      Error("illegal indirection (not a pointer)", 1);
      return ERROR;
    }
    
    if (Type__IsaFunction(type))
      return p;
    
    return (Instruction *)
      Indirection__Indirection(p, Type__Pointed_Type(type));
  }
}


/* (+ x) = x */

Instruction * make_oper_mono_plus(l, del)
    List * l;
    int del;
{
  Instruction * result;
  
  if (del) {
    if (l->info)
      DELETE((List *) l->info);
    result = 0;
  }
  else
    result = MAKE((List *) l->info);

  free(l);
  
  return result;
}


/* - */

Instruction * make_oper_mono_minus(l, del)
    List * l;
    int del;
{
  if (del) {
    if (l->info)
      DELETE((List *) l->info);
    free(l);
    return 0;
  }
  else {
    Instruction * arg = MAKE((List *) l->info);
    Type * type;
    
    free(l);
    
    if (arg == ERROR)
      return ERROR;
    
    type = GetExprType(arg);
    
    if (IsaConst (arg)) {
      /* c'est ce qui se passe lorsqu'on ecrit un nombre negatif */
      if (type == Type_Int)
	((Int *) arg)->_value = -(((Int *) arg)->_value);
      else if (type == Type_Char)
	((Char *) arg)->_value = -(((Char *) arg)->_value);
      else {
	Delete(arg);
	Error("minus pointer", 1);
	return ERROR;
      }
      
      return arg;
    }
    else {  
      Instruction ** m = (Instruction **) Malloc(sizeof(Instruction *));
      
      m[0] = arg;
      
      if ((type == Type_Int) || (type == Type_Char))
	return (Instruction *)
	  BuiltinFunctionCall__BuiltinFunctionCall
	    (Builtin_Mono_Minus, m, 1, type);
      
      /* error */
      Delete(arg);
      free(m);
      Error((type == Type_Void) ? "minus void" : "minus pointer", 1);
      return ERROR;
    }
  }
}


/* ~ */

Instruction * make_oper_complement(l, del)
    List * l;
    int del;
{
  if (del) {
    if (l->info)
      DELETE((List *) l->info);
    free(l);
    return 0;
  }
  else {
    Instruction * arg = MAKE((List *) l->info);
    Type * type;
    
    free(l);
    
    if (arg == ERROR)
      return ERROR;
    
    type = GetExprType(arg);
    
    if (IsaConst(arg)) {
      if (type == Type_Int)
	((Int *) arg)->_value = ~(((Int *) arg)->_value);
      else if (type == Type_Char)
	((Char *) arg)->_value = ~(((Char *) arg)->_value);
      else {
	Delete(arg);
	Error("pointer complement", 1);
	return ERROR;
      }
      
      return arg;
    }
    else {  
      Instruction ** m = (Instruction **) Malloc(sizeof(Instruction *));
      
      m[0] = arg;
      if ((type == Type_Int) || (type == Type_Char))
	return (Instruction *)
	  BuiltinFunctionCall__BuiltinFunctionCall
	    (Builtin_Complement, m, 1, type);
      
      /* Error */
      Delete(arg);
      free(m);
      Error((type == Type_Void) ? "void complement" : "pointer complement", 1);
      return ERROR;
    }
  }
}


Instruction * make_oper_not(l, del)
    List * l;
    int del;
{
  if (del) {
    if (l->info)
      DELETE((List *) l->info);
    free(l);
    return 0;
  }
  else {
    Instruction * arg = MAKE((List *) l->info);
    
    free(l);
    
    if (arg == ERROR)
      return ERROR;
    
    if (IsaConst(arg)) {
      if (Eval(arg)) {
	if (! IsaFunction(arg)) free(arg);
	return (Instruction *) Int__Int(0);
      }
      if (! IsaFunction(arg)) free(arg);
      return (Instruction *) Int__Int(!0);
    }
    else {
      Instruction ** m = (Instruction **) Malloc(sizeof(Instruction *));
      
      m[0] = arg;
      if (GetExprType(m[0]) != Type_Void)
	return (Instruction *)
	  BuiltinFunctionCall__BuiltinFunctionCall
	    (Builtin_Not, m, 1, Type_Int);
      
      /* Error */
      Delete(arg);
      free(m);
      Error("! void", 1);
      return ERROR;
    }
  }
}


/********************
  
  Operateurs binaires

  *******************/

/*
   fct = make_binary_oper
   info = (void **) ["opername" arg1 arg2]
*/

Instruction * make_binary_oper(l, del)
    List * l;
    int del;
{
  List ** exprs = (List **) l->info;
  
  if (del) {
    DELETE(exprs[1]);
    DELETE(exprs[2]);
    free(l);
    free(exprs[0]);
    free(exprs);
    return 0;
  }
  else {
    char * opername = (char *) exprs[0];
    Instruction ** args = (Instruction **) Malloc(2 * sizeof(Instruction *));
    Function * bfct;
    Type * valtype = 0;
    Type * t1, * t2;
    
    args[0] = MAKE(exprs[1]);
    
    if (args[0] == ERROR) {
      DELETE(exprs[2]);
      free(args);
      free(l);
      free(exprs);
      free(opername);
      return ERROR;
    }
    
    args[1] = MAKE(exprs[2]);
    if (args[1] == ERROR) {
      Delete(args[0]);
      free(args);
      free(l);
      free(exprs);
      free(opername);
      return ERROR;
    }

    t1 = GetExprType(args[0]);
    t2 = GetExprType(args[1]);
    
    free(l);
    free(exprs);
    
    bfct = find_binary_oper(opername, t1, t2, &valtype);
    free(opername);
    
    if (bfct == 0) {
      Delete(args[0]);
      Delete(args[1]);
      free(args);
      return ERROR;
    }
    
    if (IsaConst(args[0]) && IsaConst(args[1])) {
      Object val = (*(Function__BuiltinDef(bfct)))(args);
      Instruction * result = 0;
      
      Delete(args[0]);
      Delete(args[1]);
      free(args);
      
      if (valtype == Type_Int)
	result = (Instruction *) Int__Int(val);
      else if (valtype == Type_Char)
	result = (Instruction *) Char__Char(val);
      
      if (result)
	return result;
    }
    
    return (Instruction *)
      BuiltinFunctionCall__BuiltinFunctionCall(bfct, args, 2, valtype);
  }
}

/* Retourne la Function correspondant a` l'operateur et modifie valtype*/

Function * find_binary_oper(opername, t1, t2, valtype)
     char * opername;
     Type * t1;
     Type * t2;
     Type ** valtype;
{
  Function * bfct;
  
  *valtype = 0;

  if (! strcmp(opername, "==")) {
    bfct = Builtin_egal;
    *valtype = Type_Int;
  }
  else if (! strcmp(opername, "!=")) {
    bfct = Builtin_negal;
    *valtype = Type_Int;
  }
  else if (! strcmp(opername, "&&")) {
    bfct = Builtin_cond_and;
    *valtype = Type_Int;
  }
  else if (! strcmp(opername, "||")) {
    bfct = Builtin_cond_or;
    *valtype = Type_Int;
  }
  else {
    char encodedname[24];
    
    sprintf(encodedname, "%s %s ", encode_type(t1), opername);
    strcat(encodedname, encode_type(t2));

    if ((bfct = (Function *)
	 HashTable__Search(binary_oper_hashtable, (Object) encodedname))
	!= 0)
      *valtype = Function__ValType(bfct);
    else if ((bfct =
	      (Function *) HashTable__Search(binary_special_oper_hashtable,
					     (Object) encodedname))
	     != 0) {
      /* Il y a bien une fonction, mais le type retourne n'est pas fixe :
	 il est du type de celui des deux arguments qui est un pointeur,
	 a moins qu'il s'agisse de deux pointeurs auxquel cas il faut
	 verifier qu'ils soient de meme type */
      if (! Type__IsBuiltin(t1))
	if (! Type__IsBuiltin(t2))
	  if (! Type__Equal(t1, t2)) {
	    sprintf(err_msg, "%s arguments have different types", opername);
	    Error(err_msg, 1);
	    return 0;
	  }
	  else
	    *valtype = Function__ValType(bfct);
	else
	  *valtype = t1;
      else
	*valtype = t2;
    }

    if (! *valtype) {
      /* promotion de char en int */
      if (t1 == Type_Char)
	if (t2 == Type_Char)
	  return find_binary_oper(opername, Type_Int, Type_Int, valtype);
	else
	  return find_binary_oper(opername, Type_Int, t2, valtype);
      else if (t2 == Type_Char)
	return find_binary_oper(opername, t1, Type_Int, valtype);
      else {
	sprintf(err_msg, "bad argument in %s", encodedname);
	Error(err_msg, 1);
	return 0;
      }
    }
  }

  return bfct;
}

/* La meme chose, mais la fonction est un derive ou aucun
   argument n'est evalue, ceci pour les modifications */

FCT ( Object,(* find_modifbinary_oper(opername, t1, t2,	valtype)),
     	     (Object, Object, int))
     char * opername;
     Type * t1;
     Type * t2;
     Type ** valtype;
{
  Object o = (Object) find_binary_oper(opername, t1, t2, valtype);
  
  return (o)
    ? (FCT( Object,(*),(Object, Object, int)))
      HashTable__Search(binary_modifoper_hashtable, o)
    : (FCT( Object,(*),(Object, Object, int))) 0;
}
