/* ########################################################################

				switch.c

   File: switch.c
   Path: /home/fournigault/c/X11/xcoral-2.33/Smac/switch.c
   Description: 
   Created: Tue Feb 21 13:01:46 MET 1995
   Author: Bruno Pages
   Modified: Tue Feb 21 13:01:46 MET 1995
   Last maintained by: Bruno Pages

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Bruno Pages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */


#include <stdio.h>

#include "Switch_.h"
#include "switch.h"
#include "error.h"
#include "control.h"

#include "decl.h" /**/

void switch_free(List * lexprs, Instruction ** exprs, int sup)
{
  while (lexprs) {
    List * l = lexprs->next;
    
    if (lexprs->fct == make_case) {
      if (lexprs->info) {
	/* pas default: */
	DELETE((List *) lexprs->info);		/* constant : int ou char */
      }	  
    }
    else
      DELETE(lexprs);
    
    lexprs = l;  
  }
  
  if (exprs != 0) {
    int i;
    
    for (i = 0; i != sup; i += 1)
      Delete(exprs[i]);
    
    free(exprs);
  }
}
  
  
/* fct = make_switch
   info = List ** [clef, exprs]
   */

Instruction * make_switch(l, del)
    List * l;
    int del;
{
  int terminal = Is_Terminal;
  List ** m = (List **) l->info;
  Instruction * clef;
  List * lexprs;
  int iexpr, icase;
  Instruction ** exprs;
  int * cases;
  int idefault = -1;
  int prev_in_switch = in_switch;
  
  if (del) {
    DELETE(m[0]);
    
    switch_free(m[1], 0, 0);
    free(m);
    free(l);
    return 0;
    
  }
  
  Is_Terminal = 0;
  clef = MAKE(m[0]);
  
  if (clef == ERROR) {
    switch_free(m[1], 0, 0);
    free(m);
    free(l);
    return ERROR;
  }
  
  in_switch = 1;
    
#ifndef XCORAL
  if ((GetExprType(clef) != Type_Int) &&
      (GetExprType(clef) != Type_Char))
    Warning("switch key is not an int or a char");
#endif
    
  /* Compte les etiquettes et les cases */
    
  for (lexprs = m[1], iexpr = 0, icase = 0; lexprs; lexprs = lexprs->next) {
    if (lexprs->fct != make_case)
      iexpr += 1;
    else if (lexprs->info)
      /* pas default: */
      icase += 1;
    else if (idefault == -1)
      idefault = iexpr;
    else {
      Delete(clef);
      Error("more than one default: in switch", 1);
      switch_free(m[1],0, 0);
      free(m);
      free(l);
      return ERROR;
    }
  }
  
  exprs = (Instruction **) Malloc(sizeof(Instruction *) * (iexpr + 1));
  cases = (int *) Malloc((icase) ? sizeof(int) * icase * 2 : 1);
    
  /* Genere les instructions */
  
  for (lexprs = m[1], iexpr = 0, icase = 0; lexprs; ) {
    List * next = lexprs->next;
    
    if (lexprs->fct == make_case) {
      if (lexprs->info) {
	/* pas default: */
	Instruction * c =
	  MAKE((List *) lexprs->info);		/* constant : int ou char */
	
	if (c == ERROR) {
	  switch_free(next, exprs, iexpr);
	  Delete(clef);
	  free(cases);
	  free(lexprs);
	  free(m);
	  free(l);
	  return ERROR;
	}
	else {
	  int entry = Eval(c);
	  int index;
	  
	  Delete(c);
	  
	  cases[icase++] = entry;			/* clef */
	  cases[icase++] = iexpr;			/* indice dans exprs */
	  
	  for (index = 0; cases[index] != entry; index += 2)
	    ;
	  if (index != (icase - 2)) {
	    sprintf(err_msg, "duplicate case %d", entry);
	    Error(err_msg, 1);
	    switch_free(next, exprs, iexpr);
	    Delete(clef);
	    free(cases);
	    free(exprs);
	    free(m);
	    free(l);
	    return ERROR;
	  }
	}
      }
      free(lexprs);
    }
    else if ((exprs[iexpr++] = MAKE(lexprs)) == ERROR) {
      switch_free(next, exprs, iexpr);
      Delete(clef);
      free(cases);
      free(exprs);
      free(m);
      free(l);
      return ERROR;
    }
    
    lexprs = next;  
  }
  
  exprs[iexpr] = (Instruction *)
    make_control(List__List(0, 0), 0);		/* un break final */
  
  in_switch = prev_in_switch;
  
  free(m);
  free(l);
  Is_Terminal = terminal;
  
  return (Instruction *) Switch__Switch(clef, exprs, cases, icase, idefault, iexpr+1);
}

/* la fonction ne peut etre appelee que si le case est invalide */

Instruction * make_case(l, del)
    List * l;
    int del;
{
  if (del) {
    free(l);
    return 0;
  }
  
  sprintf(err_msg, "case %s out of switch", (char *) ((List *) l->info)->info);
  free(l);
  Error(err_msg, 1);
  return ERROR;
}

