/* ########################################################################

			       process.c

   File: process.c
   Path: /home/fournigault/c/X11/xcoral-2.31/process.c
   Description: 
   Created: Fri Jan 27 11:25:54 MET 1995
   Author: Lionel Fournigault
   Modified: Fri Jan 27 11:25:55 MET 1995
   Last maintained by: Lionel Fournigault

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Lionel Fournigault

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */

#include <unistd.h>

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>

#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#if STDC_HEADERS
#  include <stdlib.h>
#else
#  if HAVE_MALLOC_H
#    include <malloc.h>
#  endif
#endif

#include "xcoral.h"
#include "main_text.h"
#include "handle_key.h"
#include "process.h"
#include "text_cursor.h"
#include "page.h"
#include "main_events.h"
#include "chars_cmds.h"

/* FCT (static void, CleanFd, (Text *text) ); */
FCT (static void, GetExitShell, () );
void CleanFd();
extern int posix_openpt ();
extern int grantpt ();
extern int unlockpt ();
extern char *ptsname();

/*
**	Function name : InitShellMode
**
**	Description : 
**	Input : 
**	Ouput :
*/
void InitShellMode ()
{
    extern void ie_WR_delete ();
    int i;
    Mode *shell_mode;
    
    shell_mode = (Mode *) GetMode ("Shell");
    for ( i=32; i< 512; i++ ) { 
	shell_mode -> key [i].func = shell_key;
	shell_mode -> key [i].type = BUILTIN_FUNC;
    }
    shell_mode -> key [DELETE].func = ie_WR_delete;
    shell_mode -> key [DELETE].type = BUILTIN_FUNC;
    shell_mode -> key [CtrC].func = shell_CtrC;
    shell_mode -> key [CtrC].type = BUILTIN_FUNC;
    shell_mode -> key [CtrD].func = shell_CtrD;
    shell_mode -> key [CtrD].type = BUILTIN_FUNC;
    shell_mode -> key [CtrH].func = ie_WR_delete;
    shell_mode -> key [CtrH].type = BUILTIN_FUNC;
    shell_mode -> key [CtrM].func = shell_return;
    shell_mode -> key [CtrM].type = BUILTIN_FUNC;
    
}

/*
**	Function name : SetShellMode
**
**	Description : Passe le buffer courant en mode Shell si
**          il n'y est pas deja. Fork/exec et connection des
**          des entrees-sorties des 2 processes,
**
**	Input : Le Text courant.
**	Ouput :
*/
void SetShellMode ( text )
    Text *text;
{
    int pid, /*c_pid,*/ fdm, fds, rc;
    char *shell;
    struct termios slave_orig_term_settings;    // Saved terminal settings
    struct termios new_term_settings;   // Current terminal settings
    
    if ( strcmp(text -> current_mode -> name, "Shell") != 0 ) {
	text -> current_mode = (Mode *) GetMode ("Shell");
	text -> mwin -> mode = text -> current_mode;
	RefreshWindowMode ( text -> mwin );
	if ( (text -> current_mode -> font != text -> font)
	    && (text -> current_mode -> font != 0) )
	  RefreshWithNewFont ( text, text -> current_mode -> font );
	(void) strcpy ( text -> filename, "NoName");
    }
    else
      return;

    fdm = posix_openpt(O_RDWR); // PTY maitre
    if (fdm < 0) {
        fprintf(stderr, "Error %d on posix_openpt()\n", errno);
        return;
    }
    rc = grantpt(fdm);
    if (rc != 0) {
        fprintf(stderr, "Error %d on grantpt()\n", errno);
        return;
    }
    rc = unlockpt(fdm);
    if (rc != 0) {
        fprintf(stderr, "Error %d on unlockpt()\n", errno);
        return;
    }
    fds = open((char *)ptsname(fdm), O_RDWR); // PTY esclave

    switch ( pid = fork() ) {
    case 0: // process fils (shell) 
        // Close the master side of the PTY
        close(fdm);

        // Save the defaults parameters of the slave side of the PTY
        rc = tcgetattr(fds, &slave_orig_term_settings);

        // Set RAW mode on slave side of PTY
        new_term_settings = slave_orig_term_settings;
        cfmakeraw(&new_term_settings);
        tcsetattr(fds, TCSANOW, &new_term_settings);

        // The slave side of the PTY becomes the standard input and outputs of the child process
        close(0);               // Close standard input (current terminal)
        close(1);               // Close standard output (current terminal)
        close(2);               // Close standard error (current terminal)

        if (dup(fds) == -1)     // PTY becomes standard input (0)
	  fprintf (stderr, "dup failed\n");
        if (dup(fds) == -1)     // PTY becomes standard output (1)
	  fprintf (stderr, "dup failed\n");
        if (dup(fds) == -1)     // PTY becomes standard error (2)
	  fprintf (stderr, "dup failed\n");

        // Now the original file descriptor is useless
        close(fds);

	// Make the current process a new session leader
//        setsid();
        // As the child is a session leader, set the controlling terminal to be the slave side of the PTY
        // (Mandatory for programs like the shell to make them manage correctly their outputs)
//        ioctl(0, TIOCSCTTY, 1);
      
#if SETPGRP_VOID
      /* Linux et MacOS ok */
      (void) setpgrp ();
#else
      // freeBSD
      int c_pid;
      c_pid = getpid();
      (void) setpgrp ( c_pid, c_pid);
#endif
      shell = (char *) getenv("XCORAL_SHELL");
      if (shell == NULL) {
	// Donc bash pour Linux et MacOS, sh ou csh pour FreeBSD
	if ((access ("/bin/bash", F_OK) == 0) && (access("/bin/bash", X_OK)==0)){
	  shell = (char *) malloc(strlen("bash") + 2);
	  strcpy(shell,"bash");
	}
	else if ((access ("/bin/csh", F_OK) == 0) && (access("/bin/csh", X_OK)==0)){
	  shell = (char *) malloc(strlen("csh") + 2);
	  strcpy(shell,"csh");
	}
	else if ((access ("/bin/tcsh", F_OK) == 0) && (access("/bin/tcsh", X_OK)==0)){
	  shell = (char *) malloc(strlen("tcsh") + 2);
	  strcpy(shell,"tcsh");
	}
	else if ((access ("/bin/sh", F_OK) == 0) && (access("/bin/sh", X_OK)==0)){
	  shell = (char *) malloc(strlen("sh") + 2);
	  strcpy(shell,"sh");
	}
	else {
	  // Pas de shell utilisable, on est sur mars.
	  (void) fprintf ( stderr, "Can't find shell\n" );
	  exit(1);
	  break;
	}
      }

      if (strcmp (shell, "bash") == 0) {
	(void) execlp ( "/bin/bash", "bash", "-posix", "-i", (char *) NULL);
      }
      else if (strcmp (shell, "sh") == 0){
	 (void) execlp ( "/bin/sh", "sh", "-posix", "-i", (char *) NULL); 
      }
      else if (strcmp (shell, "csh") == 0){
	  (void) execlp( "/bin/csh", "csh", "-i", (char *) NULL);
      }
      else if (strcmp (shell, "tcsh") == 0){
	  (void) execlp( "/bin/tcsh", "tcsh", "-i", (char *) NULL);
      }
      else {
	(void) fprintf ( stderr, "exec shell failed\n" );
      }
      exit(1);
      break;
    case -1:
      perror ((char *)0);
      break;
    
    default: // process pere (xcoral)
      close (fds);
      text -> shell_id = pid;
      text -> from_shell = fdm; // standard input
      text -> to_shell = fdm;// standard output
#ifdef DEBUG      
      printf ("from_shell=%d to_shell=%d\n", text -> from_shell, text -> to_shell );
#endif
      break;
    }
    /* Pour la fonction select de la main_loop */
    SetFd ( text -> from_shell );
    text -> s_pos = 0;
    text -> s_line = 0;
}

/*
**	Function name : KillShell
**
**	Description : Cette fonction n'est appelable qu'a partir
**          du buffer courant. On peut donc tuer directement le
**          le shell associe si shell il y a.
**
**	Input : Le Text courant.
**	Ouput :
*/
void KillShell ( text )
    Text *text;
{
    if ( text -> shell_id == 0 )
	return;
    (void) kill ( text -> shell_id, SIGKILL );
    (void) GetExitShell ();
}


/*
**	Function name : GetExitShell
**
**	Description : Un shell associe a l'un des buffers
**          a ete tue soit par ^D soit sauvagement de l'exterieur.
**          Le wait doit rendre la main de suite (sinon on est mal)
**          et le zombie disparait.
**          Ensuite on cherche a savoir quel etait le buffer associe
**          a ce shell maintenant disparu et on fait le menage en
**          repassant en mode par default.
**          
**	Input : 
**	Ouput :
*/
static void GetExitShell ()
{
    int pid;
    int statusp;
    EdWin **t;
    Text *text = 0;

    pid = wait ( &statusp );
    
    for ( t = TWin; t < TWin + MAXWIN; t++ ) {
	if ( *t == 0 ) continue;
	if ( (*t) -> text -> shell_id == pid ) {
	    text = (*t) -> text;
	    break;
	}
    }

    if ( text == 0 ) {
	(void) fprintf ( stderr, "Exit shell error\n" );
	return;
    }
    
    (void) CleanFd ( text );
 
    SetDefaultMode ( text );
    TextCursorOff ( text );
    ClipOn ( text, 0 );
    InsertLines ( text , "Bye...\n",7 , 1);
    RefreshScrollBar ( dpy, text -> swin );
    TextCursorOn ( text );
    ClipOff ( text );
}


/*
**	Function name : CleanFd
**
**	Description : Mise a zero des descripteurs associes
**          au shell ( qui a disparu maintenant ).
**
**	Input : Le Text courant.
**	Ouput :
*/
void CleanFd ( text )
    Text *text;
{
    ResetFd ( text -> from_shell );
    (void) close ( text -> to_shell );
    (void) close ( text -> from_shell );
    
    text -> shell_id = 0;
    text -> to_shell = 0;
    text -> from_shell = 0;
}

/*
**	Function name : shell_return
**
**	Description : Envoie d'une commande au shell
**	Input : Le Text courant.
**	Ouput :
*/
void shell_return ( text )
    Text *text;
{
    int len;
    char *s, *to_send;

    s = (char *) GetCurrentLine ( text -> buf, &len );

    if ( text -> s_line != text -> no_current_line ) {
	text -> s_line = text -> no_current_line;
	text -> s_pos = 0;
	f_return ( text );
	if (write ( text -> to_shell, "\n", 1 ) == -1)
	  fprintf (stderr, "write failed\n");
	return;
    }

/*    if ( (s == 0) || (text -> s_pos == 0) || (len <= text -> s_pos)) {*/
    if ( (s == 0) || (len <= text -> s_pos)) {
	f_return ( text );
	if (write ( text -> to_shell, "\n", 1 ) == -1)
	  fprintf (stderr, "write failed\n");
	return;
    }

    s += text -> s_pos;
    if ( s != 0 ) {
	to_send = (char *) malloc ( (unsigned) len + 2 );
	strncpy ( to_send, s, len );
	to_send [len] =  '\n';
	to_send [len+1] = '\0';
	f_return ( text );
	if ( len == 0 ) {
	    if (write ( text -> to_shell, "\n", 1 ) == -1)
	      fprintf (stderr, "write failed\n");
	    return;
	}
	if (write ( text -> to_shell, to_send, strlen(to_send)) == -1)
	  fprintf (stderr, "write failed\n");
	(void) free ( to_send );
    }
}

/*
**	Function name : SetPosInShell
**
**	Description : Recherche de la position de la commande
**          qui sera envoyee au Shell. Cette fonction permet
**          de s'affranchir du prompt.
**    
**	Input : Le Text courant.
**	Ouput :
*/
void SetPosInShell ( text )
    Text *text;
{
    if ( text -> s_line == 0 ) { /* Fisrt time */
	text -> s_line = text -> no_current_line;
	/* La longueur du prompt */
	text -> s_pos = GetNcFromLeft ( text -> buf ); 
    }
    else {
	if ( text -> s_line != text -> no_current_line ) {
	    text -> s_line = text -> no_current_line;
	    text -> s_pos = GetNcFromLeft ( text -> buf );
	}
    }
}

/*
**	Function name : Shell_key
**
**	Description : Cette fonction est liee a toutes les touches
**          caracteres du clavier.
**	Input : 
**	Ouput :
*/
void shell_key ( text )
    Text *text;
{
    extern int last_key;

    SetPosInShell ( text );
    f_impc ( text, last_key );
    SetTextModif ( text );
}

/*
**	Function name : shell_CtrC
**
**	Description : 
**	Input : 
**	Ouput :
*/
void shell_CtrC ( text )
    Text *text;
{
    
    TextCursorOff ( text );
    ClipOn ( text, 0 );
    InsertLines ( text , "^C", 2, 0);
    RefreshScrollBar ( dpy, text -> swin );
    TextCursorOn ( text );
    ClipOff ( text );
    
#if HAVE_KILL
# if GETPGRP_VOID
    (void) kill( - getpgrp(), SIGINT );
# else
    (void) kill( - getpgrp(text -> shell_id), SIGINT );
# endif
#elif HAVE_SIGSEND
# if GETPGRP_VOID
    (void) sigsend(P_PGID, getpgrp(), SIGINT );
# else
    (void) sigsend(P_PGID, getpgrp(text -> shell_id), SIGINT );
# endif
#else
#  if GETPGRP_VOID
    (void) killpg ( getpgrp(), SIGINT );
#  else
    (void) killpg ( getpgrp(text -> shell_id), SIGINT );
#  endif
#endif
}

/*
**	Function name : shell_CtrD
**
**	Description : Envoi d'un ^D au Shell
**
**	Input : Le Text courant.
**	Ouput :
*/
void shell_CtrD ( text )
    Text *text;
{
    TextCursorOff ( text );
    ClipOn ( text, 0 );
    InsertLines ( text , "^D", 2, 0);
    RefreshScrollBar ( dpy, text -> swin );
    TextCursorOn ( text );
    ClipOff ( text );

    f_return ( text );
    if (write ( text -> to_shell, "exit\n", 5 ) == -1)
      fprintf (stderr, "write failed\n");
}

/*
**	Function name : ReadFromShell
**
**	Description : Lecture des caracteres envoyes par le
**          Shell. On essaye d'en lire le maximum d'un coup.
**
**	Input : Le Text courant.
**	Ouput :
*/
void ReadFromShell ( text )
    Text *text;
{
    int n, lines;
    char tmp [BUFSIZ];
    extern Display *dpy;

#ifdef O_NONBLOCK
    (void) fcntl ( text -> from_shell, F_SETFL, O_NONBLOCK );
#else
    (void) fcntl ( text -> from_shell, F_SETFL, O_NDELAY );
#endif

    n = read ( text -> from_shell, tmp,BUFSIZ );

    if ( n == 0 || n == -1 ) {
	/* Ya plus de shell */
 	(void) GetExitShell ();
	return;
    }

    TextCursorOff ( text );
    ClipOn ( text, 0 );
    lines = GetNewLine ( tmp, n);

    SetTextModif ( text );

    if ( lines == 0 ) {
	InsertNchar ( text -> buf, tmp, n);
/*	ClearLine ( text, text -> n1 + 1 );*/
	SetCurrentLine ( text );
    }
    else {
	InsertLines ( text, tmp, n, lines );
	RefreshScrollBar ( dpy, text -> swin );
    }

    TextCursorOn ( text );
    ClipOff ( text );
}
